@@ -1,13 +1,12 @@
 import atexit
 import glob
-import itertools
 import os
 import textwrap
 from pathlib import Path
-from typing import Dict, Iterator, List, Tuple, Union
+from typing import Any, Dict, Iterator, List, Mapping, Optional, Tuple, Union
 
 from pdm.pep517.metadata import Metadata
-from pdm.pep517.utils import is_python_package, normalize_path
+from pdm.pep517.utils import is_python_package
 
 OPEN_README = """import codecs
 
@@ -51,26 +50,49 @@ class BuildError(RuntimeError):
     pass
 
 
-def _match_path(path: str, pattern: str) -> bool:
-    return normalize_path(os.path.abspath(path)) == normalize_path(
-        os.path.abspath(pattern)
-    )
+def is_same_or_descendant_path(target: str, path: str) -> bool:
+    """Check target is same or descendant with path"""
+    try:
+        Path(target).relative_to(path)
+        return True
+    except ValueError:
+        return False
 
 
 def _merge_globs(
     include_globs: Dict[str, str], excludes_globs: Dict[str, str]
 ) -> Tuple[List[str], List[str]]:
-    includes, excludes = [], []
+    """Correctly merge includes and excludes.
+    When a pattern exists in both includes and excludes,
+    determine the priority in the following ways:
+
+    1. The one with more parts in the path wins
+    2. If part numbers are equal, the one that is more concrete wins
+    3. If both have the same part number and concrete level, *excludes* wins
+    """
+
+    def path_weight(pathname: str) -> Tuple[int, int]:
+        """Return a two-element tuple [part_num, concrete_level]"""
+        pathname_parts = Path(pathname).parts
+        wildcard_count = 0
+        if glob.has_magic(pathname):
+            for part in pathname_parts:
+                if part == "**":
+                    wildcard_count += 2
+                elif glob.has_magic(part):
+                    wildcard_count += 1
+        # The concrete level is the opposite of wildcard_count
+        return len(pathname_parts), -wildcard_count
+
+    includes = []
     for path, key in include_globs.items():
-        # The longer glob pattern wins
         if path in excludes_globs:
-            if len(key) <= len(excludes_globs[path]):
-                continue
+            if path_weight(key) <= path_weight(excludes_globs[path]):
+                continue  # Exclude wins
             else:
-                del excludes_globs[path]
+                del excludes_globs[path]  # Include wins
         includes.append(path)
-    excludes = list(excludes_globs)
-    return includes, excludes
+    return includes, list(excludes_globs)
 
 
 def _find_top_packages(root: str) -> List[str]:
@@ -103,12 +125,17 @@ def _format_dict_list(data: Dict[str, List[str]], indent: int = 4) -> str:
 class Builder:
     """Base class for building and distributing a package from given path."""
 
-    DEFAULT_EXCLUDES = ["ez_setup", "*__pycache__", "tests", "tests.*"]
+    DEFAULT_EXCLUDES = ["build"]
 
-    def __init__(self, location: Union[str, Path]) -> None:
-        self._old_cwd = None
+    def __init__(
+        self,
+        location: Union[str, Path],
+        config_settings: Optional[Mapping[str, Any]] = None,
+    ) -> None:
+        self._old_cwd: Optional[str] = None
         self.location = Path(location).absolute()
-        self._meta = None
+        self.config_settings = config_settings
+        self._meta: Optional[Metadata] = None
 
     @property
     def meta(self) -> Metadata:
@@ -123,70 +150,76 @@ class Builder:
         os.chdir(self.location)
         return self
 
-    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
+    def __exit__(self, *args: Any) -> None:
+        assert self._old_cwd
         os.chdir(self._old_cwd)
 
-    def build(self, build_dir: str, **kwargs) -> str:
+    def build(self, build_dir: str, **kwargs: Any) -> str:
         raise NotImplementedError
 
-    def _find_files_iter(self, for_sdist: bool = False) -> Iterator[str]:
+    def _get_include_and_exclude_paths(
+        self, for_sdist: bool = False
+    ) -> Tuple[List[str], List[str]]:
         includes = set()
-        find_froms = set()
-        excludes = set()
-        dont_find_froms = set()
-        source_includes = self.meta.source_includes or ["tests"]
-        meta_includes = itertools.chain(self.meta.includes, source_includes)
-        meta_excludes = list(self.meta.excludes)
+        excludes = set(self.DEFAULT_EXCLUDES)
 
+        meta_excludes = list(self.meta.excludes)
+        source_includes = self.meta.source_includes or ["tests"]
         if not for_sdist:
             # exclude source-includes for non-sdist builds
             meta_excludes.extend(source_includes)
 
-        for pat in meta_includes:
-            if os.path.basename(pat) == "*":
-                pat = pat[:-2]
-            if "*" in pat or os.path.isfile(pat):
-                includes.add(pat)
-            else:
-                find_froms.add(pat)
         if not self.meta.includes:
             top_packages = _find_top_packages(self.meta.package_dir or ".")
             if top_packages:
-                find_froms.update(top_packages)
+                includes.update(top_packages)
             else:
                 includes.add(f"{self.meta.package_dir or '.'}/*.py")
+        else:
+            includes.update(self.meta.includes)
 
-        for pat in meta_excludes:
-            if "*" in pat or os.path.isfile(pat):
-                excludes.add(pat)
-            else:
-                dont_find_froms.add(pat)
+        includes.update(source_includes)
+        excludes.update(meta_excludes)
 
-        include_globs = {path: key for key in includes for path in glob.glob(key)}
-        excludes_globs = {path: key for key in excludes for path in glob.glob(key)}
+        include_globs = {
+            os.path.normpath(path): key
+            for key in includes
+            for path in glob.iglob(key, recursive=True)
+        }
 
-        includes, excludes = _merge_globs(include_globs, excludes_globs)
-        for path in find_froms:
-            if any(_match_path(path, item) for item in dont_find_froms):
+        excludes_globs = {
+            os.path.normpath(path): key
+            for key in excludes
+            for path in glob.iglob(key, recursive=True)
+        }
+
+        include_paths, exclude_paths = _merge_globs(include_globs, excludes_globs)
+        return sorted(include_paths), sorted(exclude_paths)
+
+    def _is_excluded(self, path: str, exclude_paths: List[str]) -> bool:
+        return any(
+            is_same_or_descendant_path(path, exclude_path)
+            for exclude_path in exclude_paths
+        )
+
+    def _find_files_iter(self, for_sdist: bool = False) -> Iterator[str]:
+        includes, excludes = self._get_include_and_exclude_paths(for_sdist)
+        for include_path in includes:
+            path = Path(include_path)
+            if path.is_file():
+                yield include_path
                 continue
-            path_base = os.path.dirname(path)
-            if not path_base or path_base == ".":
-                # the path is top level itself
-                path_base = path
+            # The path is a directory name
+            for path in path.glob("**/*"):
+                if not path.is_file():
+                    continue
 
-            for root, dirs, filenames in os.walk(path):
+                rel_path = path.absolute().relative_to(self.location).as_posix()
+                if path.name.endswith(".pyc") or self._is_excluded(rel_path, excludes):
+                    continue
 
-                for filename in filenames:
-                    if filename.endswith(".pyc") or any(
-                        _match_path(os.path.join(root, filename), item)
-                        for item in excludes
-                    ):
-                        continue
-                    yield os.path.join(root, filename)
+                yield rel_path
 
-        for path in includes:
-            if os.path.isfile(path):
-                yield path
         if not for_sdist:
             return
 
@@ -194,15 +227,15 @@ class Builder:
             yield self.meta.build
 
         for pat in ("COPYING", "LICENSE"):
-            for path in glob.glob(pat + "*"):
-                if os.path.isfile(path):
-                    yield path
+            for p in glob.glob(pat + "*"):
+                if os.path.isfile(p):
+                    yield p
 
         if self.meta.readme and os.path.isfile(self.meta.readme):
             yield self.meta.readme
 
         if self.meta.filepath.exists():
-            yield "pyproject.toml"
+            yield self.meta.filepath.name
 
     def find_files_to_add(self, for_sdist: bool = False) -> List[Path]:
         """Traverse the project path and return a list of file names
@@ -304,7 +337,7 @@ class Builder:
             before="".join(before), after="".join(after), extra="".join(extra), **kwargs
         )
 
-    def format_pkginfo(self, full=True) -> str:
+    def format_pkginfo(self, full: bool = True) -> str:
         meta = self.meta
         content = METADATA_BASE.format(
             name=meta.name or "UNKNOWN",
@@ -374,7 +407,7 @@ class Builder:
         setup_py_path.write_text(self.format_setup_py(), encoding="utf-8")
 
         # Clean this temp file when process exits
-        def cleanup():
+        def cleanup() -> None:
             try:
                 setup_py_path.unlink()
             except OSError:
